from sigma.conversion.state import ConversionState
from sigma.rule import SigmaRule
from sigma.processing.pipeline import ProcessingPipeline
from sigma.conversion.base import TextQueryBackend
from sigma.conditions import ConditionItem, ConditionAND, ConditionOR, ConditionNOT, ConditionFieldEqualsValueExpression
from sigma.types import SigmaCompareExpression, SigmaRegularExpression, SigmaRegularExpressionFlag
from sigma.conversion.deferred import DeferredQueryExpression
from sigma.pipelines.cortexxdr import CortexXDR_pipeline
import sigma
import re
from typing import ClassVar, Dict, Tuple, Pattern, List, Any, Optional, Union

class CortexXDRBackend(TextQueryBackend):
    """CortexXDR backend."""

    name : ClassVar[str] = "CortexXDR backend"
    formats : Dict[str, str] = {
        "default": "Plain CortexXDR queries",        
        "json": "json output format",        
    }

    backend_processing_pipeline: ClassVar[ProcessingPipeline] = CortexXDR_pipeline()

    requires_pipeline : bool = False

    precedence : ClassVar[Tuple[ConditionItem, ConditionItem, ConditionItem]] = (ConditionNOT, ConditionAND, ConditionOR)
    parenthesize : bool = True
    group_expression : ClassVar[str] = "({expr})"   # Expression for precedence override grouping as format string with {expr} placeholder

    # Generated query tokens
    token_separator : str = " "     # separator inserted between all boolean operators
    or_token : ClassVar[str] = "or"
    and_token : ClassVar[str] = "and"
    not_token : ClassVar[str] = "not"
    eq_token : ClassVar[str] = " = "  # Token inserted between field and value (without separator)

    # String output
    ## Fields
    ### Quoting
    field_quote : ClassVar[str] = "'"                               # Character used to quote field characters if field_quote_pattern matches (or not, depending on field_quote_pattern_negation). No field name quoting is done if not set.
    field_quote_pattern : ClassVar[Pattern] = re.compile("^\\w+$")   # Quote field names if this pattern (doesn't) matches, depending on field_quote_pattern_negation. Field name is always quoted if pattern is not set.
    field_quote_pattern_negation : ClassVar[bool] = True            # Negate field_quote_pattern result. Field name is quoted if pattern doesn't matches if set to True (default).

    ### Escaping
    field_escape : ClassVar[str] = "\\"               # Character to escape particular parts defined in field_escape_pattern.
    field_escape_quote : ClassVar[bool] = True        # Escape quote string defined in field_quote
    field_escape_pattern : ClassVar[Pattern] = re.compile("\\s")   # All matches of this pattern are prepended with the string contained in field_escape.

    ## Values
    str_quote         : ClassVar[str] = '"'     # string quoting character (added as escaping character)
    str_quote_pattern : ClassVar[Optional[Pattern]] = re.compile("^ENUM\.[A-Z_]{1,}$")      # Quote string values that match (or don't match) this pattern
    str_quote_pattern_negation : ClassVar[bool] = True  # Negate str_quote_pattern result
    escape_char       : ClassVar[str] = "\\"    # Escaping character for special characters inside string
    wildcard_multi    : ClassVar[str] = "*"     # Character used as multi-character wildcard
    wildcard_single   : ClassVar[str] = "*"     # Character used as single-character wildcard
    add_escaped       : ClassVar[str] = "\\"    # Characters quoted in addition to wildcards and string quote
    filter_chars      : ClassVar[str] = ""      # Characters filtered
    bool_values       : ClassVar[Dict[bool, str]] = {   # Values to which boolean values are mapped.
        True: "true",
        False: "false",
    }

    # String matching operators. if none is appropriate eq_token is used.
    startswith_expression : ClassVar[str] = "{field} contains {value}"
    endswith_expression   : ClassVar[str] = "{field} contains {value}"
    contains_expression   : ClassVar[str] = "{field} contains {value}"
    wildcard_match_expression : ClassVar[str] = "contains"      # Special expression if wildcards can't be matched with the eq_token operator

    # Regular expressions
    # Regular expression query as format string with placeholders {field}, {regex}, {flag_x} where x
    # is one of the flags shortcuts supported by Sigma (currently i, m and s) and refers to the
    # token stored in the class variable re_flags.
    re_expression : ClassVar[str] = "{field} ~= \"{regex}\""
    re_escape_char : ClassVar[str] = "\\"               # Character used for escaping in regular expressions
    re_escape : ClassVar[Tuple[str]] = ()               # List of strings that are escaped
    re_escape_escape_char : bool = True                 # If True, the escape character is also escaped
    re_flag_prefix : bool = True                        # If True, the flags are prepended as (?x) group at the beginning of the regular expression, e.g. (?i). If this is not supported by the target, it should be set to False.
    # Mapping from SigmaRegularExpressionFlag values to static string templates that are used in
    # flag_x placeholders in re_expression template.
    # By default, i, m and s are defined. If a flag is not supported by the target query language,
    # remove it from re_flags or don't define it to ensure proper error handling in case of appearance.
    #re_flags : Dict[SigmaRegularExpressionFlag, str] = {
    #    SigmaRegularExpressionFlag.IGNORECASE: "i",
    #    SigmaRegularExpressionFlag.MULTILINE : "m",
    #    SigmaRegularExpressionFlag.DOTALL    : "s",
    #}

    # cidr expressions
    cidr_wildcard : ClassVar[str] = "*"    # Character used as single wildcard
    cidr_expression : ClassVar[str] = "{field} incidr \"{value}\""    # CIDR expression query as format string with placeholders {field} = {value}

    # Numeric comparison operators
    compare_op_expression : ClassVar[str] = "{field} {operator} {value}"  # Compare operation query as format string with placeholders {field}, {operator} and {value}
    # Mapping between CompareOperators elements and strings used as replacement for {operator} in compare_op_expression
    compare_operators : ClassVar[Dict[SigmaCompareExpression.CompareOperators, str]] = {
        SigmaCompareExpression.CompareOperators.LT  : "<",
        SigmaCompareExpression.CompareOperators.LTE : "<=",
        SigmaCompareExpression.CompareOperators.GT  : ">",
        SigmaCompareExpression.CompareOperators.GTE : ">=",
    }

    # Expression for comparing two event fields
    field_equals_field_expression : ClassVar[Optional[str]] = None  # Field comparison expression with the placeholders {field1} and {field2} corresponding to left field and right value side of Sigma detection item
    field_equals_field_escaping_quoting : Tuple[bool, bool] = (True, True)   # If regular field-escaping/quoting is applied to field1 and field2. A custom escaping/quoting can be implemented in the convert_condition_field_eq_field_escape_and_quote method.

    # Null/None expressions
    #field_null_expression : ClassVar[str] = "{field} is null"          # Expression for field has null value as format string with {field} placeholder for field name

    # Field existence condition expressions.
    field_exists_expression : ClassVar[str] = '{field} contains "*"'             # Expression for field existence as format string with {field} placeholder for field name
    field_not_exists_expression : ClassVar[str] = '{field} not contains "*"'      # Expression for field non-existence as format string with {field} placeholder for field name. If not set, field_exists_expression is negated with boolean NOT.

    # Field value in list, e.g. "field in (value list)" or "field containsall (value list)"
    convert_or_as_in : ClassVar[bool] = True                     # Convert OR as in-expression
    convert_and_as_in : ClassVar[bool] = False                    # Convert AND as in-expression
    in_expressions_allow_wildcards : ClassVar[bool] = True       # Values in list can contain wildcards. If set to False (default) only plain values are converted into in-expressions.
    field_in_list_expression : ClassVar[str] = "{field} {op} ({list})"  # Expression for field in list of values as format string with placeholders {field}, {op} and {list}
    or_in_operator : ClassVar[str] = "in"               # Operator used to convert OR into in-expressions. Must be set if convert_or_as_in is set
    list_separator : ClassVar[str] = ", "               # List element separator

    # Value not bound to a field
    #unbound_value_str_expression : ClassVar[str] = '"{value}"'   # Expression for string value not bound to a field as format string with placeholder {value}
    #unbound_value_num_expression : ClassVar[str] = '{value}'     # Expression for number value not bound to a field as format string with placeholder {value}
    #unbound_value_re_expression : ClassVar[str] = '_=~{value}'   # Expression for regular expression not bound to a field as format string with placeholder {value} and {flag_x} as described for re_expression

    # Query finalization: appending and concatenating deferred query part
    deferred_start : ClassVar[str] = "\n| "               # String used as separator between main query and deferred parts
    deferred_separator : ClassVar[str] = "\n| "           # String used to join multiple deferred query parts
    #deferred_only_query : ClassVar[str] = "*"            # String used as query if final query only contains deferred expression

    def finalize_query_default(self, rule: SigmaRule, query: str, index: int, state: ConversionState) -> Any:
        """
        Finalize conversion result of a query
        This will add the dataset name to the beginning of the query based on the rule category.
        The dataset name is generated by the ProcessingPipeline state which is why the pipeline is automatically applied.
        """
        index_type, index = state.processing_state.get('dataset_preset', 'dataset::xdr_dataset').split('::')

        full_query = f'{index_type}={index} | filter {query}'
        full_query += f' | fields {",".join(rule.fields)}' if rule.fields else ''

        return full_query

    def finalize_output_default(self, queries: List[str]) -> Any:
        return queries
    
    def finalize_query_json(self, rule: SigmaRule, query: str, index: int, state: ConversionState) -> Any:
        index_type, index = state.processing_state.get('dataset_preset', 'dataset::xdr_dataset').split('::')

        full_query = f'{index_type}={index} | filter {query}'
        full_query += f' | fields {",".join(rule.fields)}' if rule.fields else ''

        return {"query":full_query, "title":rule.title, "id":rule.id, "description": rule.description}

    def finalize_output_json(self, queries: List[str]) -> Any:
        return {"queries":queries}


from .cortexxdr import CortexXDRBackend
# TODO: add all backend classes that should be exposed to the user of your backend in the import statement above.

backends = {        # Mapping between backend identifiers and classes. This is used by the pySigma plugin system to recognize backends and expose them with the identifier.
    "cortexxdr": CortexXDRBackend,
}


from .cortexxdr import CortexXDR_pipeline
# TODO: add all pipelines that should be exposed to the user of your backend in the import statement above.

pipelines = {
    "CortexXDR_pipeline": CortexXDR_pipeline,   # TODO: adapt identifier to something approproiate
}


from typing import Union
from sigma.pipelines.common import logsource_windows, windows_logsource_mapping
from sigma.processing.transformations import ConditionTransformation, AddConditionTransformation, FieldMappingTransformation, DetectionItemFailureTransformation, RuleFailureTransformation, ChangeLogsourceTransformation, SetStateTransformation
from sigma.processing.conditions import LogsourceCondition, ExcludeFieldCondition, RuleProcessingItemAppliedCondition
from sigma.processing.pipeline import QueryPostprocessingItem, ProcessingItem, ProcessingPipeline
from sigma.processing.postprocessing import QueryPostprocessingTransformation
from sigma.rule import SigmaDetectionItem, SigmaDetection, SigmaRule
from sigma.exceptions import SigmaTransformationError
import re
import json

class InvalidFieldTransformation(DetectionItemFailureTransformation):
    """
    Overrides the apply_detection_item() method from DetectionItemFailureTransformation to also include the field name
    in the error message
    """

    def apply_detection_item(self, detection_item: SigmaDetectionItem) -> None:
        field_name = detection_item.field
        self.message = f"Invalid SigmaDetectionItem field name encountered: {field_name}. " + self.message
        raise SigmaTransformationError(self.message)

## Custom QueryPostprocessingTransformation to convert string values in IntegrityLevel field to integer range, if applicable
class ReplaceIntegrityLevelQueryTransformation(QueryPostprocessingTransformation):
    """Replace query part specified by regular expression with a given string."""

    def apply(
        self, pipeline: "sigma.processing.pipeline.ProcessingPipeline", rule: SigmaRule, query: Union[str, dict]
    ) -> Union[str, dict]:

        if isinstance(query, dict):
            output_type = 'json'
            query = json.dumps(query)
        else:
            output_type = 'default'

        self.identifier = 'replace_integrity_thing'
        field_name = 'action_process_integrity_level'

        super().apply(pipeline, rule, query)

        integrity_level_ranges ={
            'UNTRUSTED': f'{field_name} lt 4096',
            'LOW': f'({field_name} gte 4096 and {field_name} lt 8192)',
            'MEDIUM': f'({field_name} gte 8192 and {field_name} lt 12288)',
            'HIGH': f'({field_name} gte 12288 and {field_name} lt 16384)',
            'SYSTEM': f'{field_name} gte 16384'
        }

        single_pattern = '(?i)' + field_name + ' = "(' + '|'.join(integrity_level_ranges) + ')"'
        multi_pattern = '(?i)' + field_name + " in \\(((\"(" + '|'.join(integrity_level_ranges) + ")\")((, )*)){1,}\\)"

        if re.search(single_pattern, query): # for single value
            for level in integrity_level_ranges.items():
                query = re.sub(f'(?i){field_name} = "{level[0]}"', level[1], query)

        while re.search(multi_pattern, query): # for multiple values
            matches = re.search(multi_pattern, query)
            target_string = matches.group(0)

            values = (re.sub(f"(?i){field_name} in \\(", '', target_string)).replace(')', '').replace('"', '').split(',')
            replacement_values = []

            for value in values:
                if value.strip().upper() in integrity_level_ranges.keys():
                    replacement_values.append(integrity_level_ranges[value.strip().upper()])

            replacement_string = '(' + ' or '.join(replacement_values) + ')'
            query = query.replace(target_string, replacement_string)

        if output_type == 'json':
            query = json.loads(query)

        return query


def CortexXDR_pipeline() -> ProcessingPipeline:

    logsource_category_to_event_type = {
        'process_creation':{
            "event_type": "ENUM.PROCESS",
            "event_sub_type": "ENUM.PROCESS_START"
        },
        "file_change": {
            "event_type":"ENUM.FILE"
        },
        "file_rename": {
            "event_type":"ENUM.FILE",
            'event_sub_type': 'ENUM.FILE_RENAME'
        },
        "file_delete": {
            "event_type":"ENUM.FILE",
            'event_sub_type': 'ENUM.FILE_REMOVE'
        },
        "file_event":{
            "event_type":"ENUM.FILE"
        },
        'image_load':{
            "event_type":"ENUM.LOAD_IMAGE"
        },
        "registry_add":{
            "event_type":"ENUM.REGISTRY",
            'event_sub_type': 'ENUM.REGISTRY_CREATE_KEY'
        },
        "registry_delete":{
            "event_type":"ENUM.REGISTRY",
            'event_sub_type': ['ENUM.REGISTRY_DELETE_KEY', 'ENUM.REGISTRY_DELETE_VALUE']
        },
        "registry_event":{
            "event_type":"ENUM.REGISTRY"
        },
        "registry_set":{
            "event_type":"ENUM.REGISTRY",
            'event_sub_type': 'ENUM.REGISTRY_SET_VALUE'
        },
        "network_connection": {
            "event_type":"ENUM.NETWORK"
        },
        "firewall":{
            "event_type":"ENUM.NETWORK"
        }
    }

    generic_translation_dict = {
        'User': 'actor_effective_username',
        'CommandLine': 'actor_process_image_command_line',
        'Image': 'actor_process_image_path',
        "LogonId": "actor_process_logon_id",
        "Product": "actor_process_signature_product",
        "Company": "actor_process_signature_vendor",
        "IntegrityLevel": "actor_process_integrity_level",
        'CurrentDirectory': 'actor_process_cwd',
        'ProcessId': 'actor_process_os_id',
        'ParentProcessId': 'causality_actor_process_os_id',
        'ParentCommandLine': 'causality_actor_process_command_line',
        'ParentImage': 'causality_actor_process_image_path',
        "ParentUser": "causality_actor_effective_username",
        'ParentIntegrityLevel': 'causality_actor_process_integrity_level',
        'ParentLogonId': 'causality_actor_process_logon_id',
        'ParentProduct': 'causality_actor_process_signature_product',
        'ParentCompany': 'causality_actor_process_signature_vendor'
        #'ProcessGuid': ?,
        #'ParentProcessGuid': ?,
        #'FileVersion': ?,
        #'Description': ?,
        #'OriginalFileName': ?,
        #'EventType': ?,
    }

    translation_dict = {
        'process':{
            'index': {
                'name': 'xdr_process',
                'type': 'preset'
            },
            'category': ['process_creation'],
            'fields':{
                "User":"action_process_username",
                "CommandLine":"action_process_image_command_line",
                "Image":"action_process_image_path",
                "LogonId": "action_process_logon_id",
                "Product":"action_process_signature_product",
                "Company":"action_process_signature_vendor",
                "IntegrityLevel":"action_process_integrity_level",
                "CurrentDirectory":"action_process_cwd",
                "ProcessId":"action_process_os_pid",
                "ParentProcessId":"actor_process_os_pid",
                "ParentCommandLine":"actor_process_image_command_line",
                "ParentImage":"actor_process_image_path",
                "ParentUser": "actor_effective_username",
                'ParentIntegrityLevel': 'actor_process_integrity_level',
                'ParentLogonId': 'actor_process_logon_id',
                'ParentProduct': 'actor_process_signature_product',
                'ParentCompany': 'actor_process_signature_vendor',
                "md5":"action_process_image_md5",
                "sha256":"action_process_image_sha256",
                #"ProcessGuid": ?,
                #'ParentProcessGuid': ?,
                #"FileVersion": ?,
                #"Description": ?,
                #"OriginalFileName": ?,
                #"Hashes": ?,
                #'EventType': ?,
                #"LogonGuid": ?,
                #"TerminalSessionId": ?,
                #"sha1": ?,
            }
        },
        'file':{
            'index': {
                'name': 'xdr_file',
                'type': 'preset'
            },
            'category': ['file_change','file_rename','file_delete','file_event'],
            'fields':{
                **generic_translation_dict,
                'TargetFilename': 'action_file_name',
                'SourceFilename': 'action_file_previous_file_name',
                #'sha1': ?,
                #'sha256': ?,
                #'md5': ?,
                #'Hashes': ?,
                #'CreationUtcTime': ?
            }
        },
        'image_load':{
            'index': {
                'name': 'xdr_image_load',
                'type': 'preset'
            },
            'category': ['image_load'],
            'fields':{
                **generic_translation_dict,
                'ImageLoaded': 'action_module_path',
                'md5': 'action_module_md5',
                'sha256': 'action_module_sha256',
                #'sha1': ?,
                #'Signed': ?,
                #'Signature': ?,
                #'SignatureStatus': ?
            }
        },
        "registry":{
            'index': {
                'name': 'xdr_registry',
                'type': 'preset'
            },
            'category': ['registry_add', 'registry_delete', 'registry_event', 'registry_set'],
            'fields': {
                **generic_translation_dict,
                'TargetObject': 'action_registry_key_name',
                'Details': ['action_registry_value_name', 'action_registry_data'],
                #'NewName': ?,
            }
        },
        'network':{
            'index': {
                'name': 'network_story', # or xdr_agent_network - not sure the difference
                'type': 'preset'
            },
            'category': ['network_connection','firewall'],
            'fields': {
                **generic_translation_dict,
                # Have to have local/remote like this because direction isn't defined that I can tell
                'DestinationPort': ['action_local_port', 'action_remote_port'],
                'DestinationIp': ['action_local_ip', 'action_remote_ip'],
                'SourcePort': ['action_local_port', 'action_remote_port'],
                'SourceIp': ['action_local_ip', 'action_remote_ip'],
                'Protocol': 'action_network_protocol',
                'dst_ip': ['action_local_ip', 'action_remote_ip'],
                'dst_port': ['action_local_port', 'action_remote_port'],
                'src_ip': ['action_local_ip', 'action_remote_ip'],
                'src_port': ['action_local_port', 'action_remote_port'],
                'DestinationHostname': 'action_external_hostname',
                'SourceHostname': 'agent_hostname',
                #'Initiated': ?,
                #'SourceIsIpv6': ?,
                #'SourcePortName': ?,
                #'DestinationIsIpv6': ?,
                #'DestinationPortName': ?
            }
        }
    }

    os_translation_dict = {
        'windows':{ 'agent_os_type': 'ENUM.AGENT_OS_WINDOWS' },
        'linux':{ 'agent_os_type': 'ENUM.AGENT_OS_LINUX' },
        'macos':{ 'agent_os_type': 'ENUM.AGENT_OS_MAC' }
    }

    os_filter = [
        ProcessingItem(
            identifier=f"cortexxdr_{os_name}_os",
            transformation=AddConditionTransformation(translation_value),
            rule_conditions=[
                LogsourceCondition(product=os_name)
            ]
        )
        for os_name, translation_value in os_translation_dict.items()
    ]

    event_type_filters = [
        ProcessingItem(
            identifier=f"cortex_{event_type}_eventtype",
            transformation=AddConditionTransformation(translation_value),
            rule_conditions = [
                LogsourceCondition(category=event_type)
            ]
        )
        for event_type, translation_value in logsource_category_to_event_type.items()
    ]

    dataset_preset_configuration = [
        ProcessingItem(
            identifier=f"cortex_dataset_preset_{activity_type}_config",
            transformation=SetStateTransformation('dataset_preset', details['index']['type'] + '::' + details['index']['name']),
            rule_conditions=[
                LogsourceCondition(category=category)
                for category in details['category']
            ],
            rule_condition_linking=any
        )
        for activity_type, details in translation_dict.items()
    ]

    field_mappings = [
        ProcessingItem(
            identifier=f"cortex_{activity_type}_fieldmapping",
            transformation=FieldMappingTransformation(details['fields']),
            rule_conditions=[
                LogsourceCondition(category=category)
                for category in details['category']
            ],
            rule_condition_linking=any
        )
        for activity_type, details in translation_dict.items()
    ]

    change_logsource_info = [
        # Add service to be SentinelOne for pretty much everything
        ProcessingItem(
            identifier="cortex_logsource",
            transformation=ChangeLogsourceTransformation(
                service="cortex"
            ),
            rule_condition_linking=any,
            rule_conditions=[
                LogsourceCondition(category=category)
                for category in logsource_category_to_event_type.keys()
            ]
        ),
    ]

    unsupported_rule_types = [
        # Show error if unsupported option
        ProcessingItem(
            identifier="cortex_fail_rule_not_supported",
            rule_condition_linking=any,
            transformation=RuleFailureTransformation("Rule type not yet supported by the Cortex XDR Sigma backend"),
            rule_condition_negation=True,
            rule_conditions=[
                RuleProcessingItemAppliedCondition("cortex_logsource")
            ]
        )
    ]
    
    unsupported_field_name_by_category = [
        ProcessingItem(
            identifier=f'cortex_fail_field_not_supported_{activity_type}',
            transformation=InvalidFieldTransformation("This pipeline only supports the following fields for " + activity_type + " activity:\n{" + 
            '}, {'.join(details['fields'].keys()) + '}'),
            field_name_conditions=[
                ExcludeFieldCondition(fields=details['fields'].keys())
            ],
            rule_conditions=[
                LogsourceCondition(category=category)
                for category in details['category']
            ],
            rule_condition_linking=any
        )
        for activity_type, details in translation_dict.items()
    ]

    return ProcessingPipeline(
        name="CortexXDR pipeline",
        priority=50,
        items = [
            *unsupported_field_name_by_category,
            *dataset_preset_configuration,
            *os_filter, 
            *event_type_filters,
            *field_mappings,
            *change_logsource_info,
            *unsupported_rule_types,
        ],
        postprocessing_items=[
            QueryPostprocessingItem(
                identifier=f'replace_integrity_level',
                transformation=ReplaceIntegrityLevelQueryTransformation(),
            ),
        ],
    )
